'use strict';

Object.defineProperty(exports, '__esModule', {
  value: true
});

var _createClass = (function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ('value' in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; })();

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { 'default': obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError('Cannot call a class as a function'); } }

require('native-promise-only');

var _immutableContribCursor = require('immutable/contrib/cursor');

var _immutableContribCursor2 = _interopRequireDefault(_immutableContribCursor);

var _vertebraeEvents = require('vertebrae/events');

var _vertebraeEvents2 = _interopRequireDefault(_vertebraeEvents);

var _immstruct = require('immstruct');

var _immstruct2 = _interopRequireDefault(_immstruct);

var _immutable = require('immutable');

var _immutable2 = _interopRequireDefault(_immutable);

var _vertebraeUtils = require('vertebrae/utils');

var _storeCreateCacheKey = require('./store/create-cache-key');

var _storeCreateCacheKey2 = _interopRequireDefault(_storeCreateCacheKey);

var Store = (function () {
  function Store() {
    var _this = this;

    var stores = arguments[0] === undefined ? {} : arguments[0];
    var options = arguments[1] === undefined ? {} : arguments[1];

    _classCallCheck(this, Store);

    // store cookie for server side API request authentication
    if (options.cookie) {
      this.cookie = options.cookie;
    }

    // instantly resolve "fetch" promise, for non-blocking rendering on client side
    // use "fetch" event to re-render when this is enabled
    this.instantResolve = false;
    if (typeof options.instantResolve !== 'undefined') {
      this.instantResolve = options.instantResolve;
    }

    // automatically clear caches after mutation operations, default true
    this.autoClearCaches = true;
    if (typeof options.autoClearCaches !== 'undefined') {
      this.autoClearCaches = options.autoClearCaches;
    }

    // allowed status codes
    // return empty store with these status codes
    // this allows us to show proper data for logged-in users
    // and prevents error for users who are not logged-in
    this.allowedStatusCodes = [401, 403];
    if (Array.isArray(options.allowedStatusCodes)) {
      this.allowedStatusCodes = options.allowedStatusCodes;
    }

    // used for storing all collection/model data, fetch returns
    // fresh cursors to this structure
    this.cached = new _immstruct2['default'].Immstruct().get();

    // used for tracking stale caches, it's better to mark caches stale
    // and re-fetch than mutating the cache and causing glitchy re-renders
    this.staleCaches = [];

    // used for tracking ongoing fetch requests to prevent multiple
    // concurrent fetch calls to same API
    this.ongoingFetches = [];

    // empty instances of stores
    // clone actual instances from these
    this.stores = {};

    Object.keys(stores).forEach(function (storeId) {
      var store = stores[storeId];
      _this.stores[storeId] = new store();
      _this.cached.cursor().set(storeId, _immutable2['default'].fromJS({}));
    });

    // allows Store to be used as event bus
    (0, _vertebraeUtils.extend)(this, _vertebraeEvents2['default']);
  }

  _createClass(Store, [{
    key: 'dispatch',
    value: function dispatch(action, storeId, storeOptions, attrs) {
      return this[action].call(this, storeId, storeOptions, attrs);
    }
  }, {
    key: 'create',
    value: function create(storeId, storeOptions, attrs) {
      var _this2 = this;

      return new Promise(function (resolve, reject) {
        if (!attrs) {
          attrs = storeOptions;
          storeOptions = {};
        }

        var store = _this2.get(storeId);
        store.storeOptions = (0, _vertebraeUtils.extend)({}, storeOptions);
        var cacheKey = (0, _storeCreateCacheKey2['default'])(store);

        if (typeof store.create !== 'function') {
          _this2.trigger('create:' + storeId, new Error('You can call create only for collections!'));
          return;
        }

        // optimistic create, save previous value for error scenario
        var previousCollection = _this2.cached.cursor([storeId, cacheKey]).deref();
        _this2.cached.cursor([storeId, cacheKey]).update(function (collection) {
          var newItem = _immutable2['default'].fromJS(attrs);
          if (collection) {
            return collection.push(newItem);
          } else {
            return new _immutable2['default'].List([newItem]);
          }
        });

        store.create(attrs, {
          success: function success(model, response) {
            if (_this2.autoClearCaches) {
              _this2.clearCache(storeId, store.storeOptions);
            }
            _this2.trigger('create:' + storeId, null, {
              cacheKey: cacheKey,
              store: storeId,
              options: storeOptions,
              result: model
            });
            resolve(model);
          },
          error: function error(model, response) {
            var error = new Error('Creating new item to store \'' + storeId + '\' failed');
            error.store = storeId;
            error.result = response;
            error.options = storeOptions;
            _this2.cached.cursor([storeId, cacheKey]).update(function () {
              return previousCollection;
            });
            _this2.trigger('create:' + storeId, error);
            reject(error);
          }
        });
      });
    }
  }, {
    key: 'update',
    value: function update(storeId, storeOptions, attrs) {
      var _this3 = this;

      return new Promise(function (resolve, reject) {
        var store = _this3.get(storeId);
        store.storeOptions = (0, _vertebraeUtils.extend)({}, storeOptions);
        var cacheKey = (0, _storeCreateCacheKey2['default'])(store);

        if (typeof store.save !== 'function') {
          _this3.trigger('update:' + storeId, new Error('You can call update only for models!'));
          return;
        }

        // optimistic update, save previous value for error scenario
        var previousModel = _this3.cached.cursor([storeId, cacheKey]).deref();
        if (previousModel) {
          _this3.cached.cursor([storeId, cacheKey]).merge(attrs);
        }

        store.save(attrs, {
          success: function success(model, response) {
            if (_this3.autoClearCaches) {
              _this3.clearCache(storeId, store.storeOptions);
            }
            _this3.trigger('update:' + storeId, null, {
              cacheKey: (0, _storeCreateCacheKey2['default'])(store),
              store: storeId,
              options: storeOptions,
              result: model
            });
            resolve(model);
          },
          error: function error(model, response) {
            var error = new Error('Updating \'' + storeId + '\' failed');
            error.store = storeId;
            error.result = response;
            error.options = storeOptions;
            if (previousModel) {
              _this3.cached.cursor([storeId, cacheKey]).merge(previousModel);
            }
            _this3.trigger('update:' + storeId, error);
            reject(error);
          }
        });
      });
    }
  }, {
    key: 'delete',
    value: function _delete(storeId, storeOptions) {
      var _this4 = this;

      return new Promise(function (resolve, reject) {
        var store = _this4.get(storeId);
        store.storeOptions = (0, _vertebraeUtils.extend)({}, storeOptions);
        // Model#destroy needs id attribute or it considers model new and triggers success callback straight away
        store.set('id', store.storeOptions.id);
        var cacheKey = (0, _storeCreateCacheKey2['default'])(store);

        if (typeof store.destroy !== 'function') {
          _this4.trigger('delete:' + storeId, new Error('You can call destroy only for models!'));
          return;
        }

        // optimistic delete, save previous value for error scenario
        var previousModel = _this4.cached.cursor([storeId, cacheKey]).deref();
        _this4.cached.cursor(storeId)['delete'](cacheKey);

        store.destroy({
          success: function success(model, response) {
            if (_this4.autoClearCaches) {
              _this4.clearCache(storeId, store.storeOptions);
            }
            _this4.trigger('delete:' + storeId, null, {
              cacheKey: (0, _storeCreateCacheKey2['default'])(store),
              store: storeId,
              options: storeOptions,
              result: model
            });
            resolve(model);
          },
          error: function error(model, response) {
            var error = new Error('Deleting \'' + storeId + '\' failed');
            error.store = storeId;
            error.result = response;
            error.options = storeOptions;
            _this4.cached.cursor(storeId).set(cacheKey, previousModel);
            _this4.trigger('delete:' + storeId, error);
            reject(error);
          }
        });
      });
    }
  }, {
    key: 'expire',
    value: function expire(storeId, storeOptions) {
      var _this5 = this;

      return new Promise(function (resolve, reject) {
        var store = _this5.get(storeId);
        store.storeOptions = (0, _vertebraeUtils.extend)({}, storeOptions);
        _this5.clearCache(storeId, store.storeOptions);
        _this5.trigger('expire:' + storeId, null, {
          cacheKey: (0, _storeCreateCacheKey2['default'])(store),
          store: storeId,
          options: storeOptions
        });
        return resolve(storeId);
      });
    }
  }, {
    key: 'clearCache',
    value: function clearCache(storeId, storeOptions) {
      var _this6 = this;

      var relatedCaches = {};
      var relatedCacheKeys = [];

      var store = this.get(storeId);
      if (!store) {
        return;
      }
      store.storeOptions = storeOptions;
      var cacheKey = (0, _storeCreateCacheKey2['default'])(store);

      if (cacheKey) {
        // mark cache as stale
        if (this.cached.cursor().getIn([storeId, cacheKey])) {
          this.staleCaches.push({ id: storeId, key: cacheKey });
        }
      } else {
        // mark all caches for store stale if cacheKey doesn't exist
        Object.keys(this.cached.cursor().getIn([storeId]).toJS()).forEach(function (key) {
          _this6.staleCaches.push({ id: storeId, key: key });
        });
      }

      // mark related caches as stale
      if (typeof store.relatedCaches === 'function') {
        relatedCaches = store.relatedCaches();
        relatedCacheKeys = Object.keys(relatedCaches);

        if (relatedCacheKeys.length) {
          relatedCacheKeys.forEach(function (id) {
            var key = relatedCaches[id];
            if (_this6.cached.cursor().getIn([id, key])) {
              _this6.staleCaches.push({ id: id, key: key });
            }
          });
        }
      }
    }
  }, {
    key: 'clearCookie',
    value: function clearCookie() {
      this.cookie = null;
    }
  }, {
    key: 'bootstrap',

    // bootstrap caches from JSON snapshot
    value: function bootstrap(json) {
      if (!json) {
        return;
      }

      return this.cached.cursor().update(function () {
        return _immutable2['default'].fromJS(JSON.parse(json));
      });
    }
  }, {
    key: 'snapshot',

    // export snapshot of current caches to JSON
    value: function snapshot() {
      return JSON.stringify(this.cached.cursor().toJSON());
    }
  }, {
    key: 'get',

    // returns cloned empty store instance
    value: function get(storeId) {
      var store = this.stores[storeId];
      return store ? store.clone() : null;
    }
  }, {
    key: 'isCacheStale',

    // checks if cache has gone stale
    value: function isCacheStale(id, key) {
      return this.staleCaches.some(function (cache) {
        return cache.id === id && cache.key === key;
      });
    }
  }, {
    key: 'markCacheFresh',

    // mark cache as not stale
    value: function markCacheFresh(id, key) {
      this.staleCaches = this.staleCaches.filter(function (cache) {
        return cache.id !== id && cache.key !== key;
      });
    }
  }, {
    key: 'ongoingFetch',
    value: function ongoingFetch(id, key) {
      return this.ongoingFetches.filter(function (fetch) {
        return fetch.id === id && fetch.key === key.toString();
      })[0];
    }
  }, {
    key: 'markFetchOngoing',
    value: function markFetchOngoing(id, key, promise) {
      this.ongoingFetches.push({
        id: id,
        key: key,
        promise: promise
      });
    }
  }, {
    key: 'markFetchCompleted',
    value: function markFetchCompleted(id, key) {
      this.ongoingFetches = this.ongoingFetches.filter(function (fetch) {
        return fetch.id !== id && fetch.key !== key;
      });
    }
  }, {
    key: 'createFetchOptions',
    value: function createFetchOptions() {
      var fetchOptions = {};
      if (this.cookie) {
        fetchOptions.headers = { 'cookie': this.cookie };
      }
      return fetchOptions;
    }
  }, {
    key: 'createStoreInstance',
    value: function createStoreInstance(storeId, options) {
      var store = this.get(storeId);
      if (!store) {
        throw new Error('Store ' + storeId + ' not registered.');
      }
      store.storeOptions = (0, _vertebraeUtils.extend)({}, options);
      return store;
    }
  }, {
    key: 'fetch',

    // get store from cache or fetch from server
    // TODO: split to smaller parts, this is really complicated to comprehend
    value: function fetch(storeId, options) {
      var _this7 = this;

      return new Promise(function (resolve, reject) {
        var fetchOptions = _this7.createFetchOptions();
        var instantResolve = _this7.instantResolve;
        var store = _this7.createStoreInstance(storeId, options);
        var cacheKey = (0, _storeCreateCacheKey2['default'])(store);
        if (!cacheKey) {
          reject(new Error('Store ' + storeId + ' has no cacheKey method.'));
        }
        var cachedStore = _this7.cached.cursor([storeId, cacheKey]);
        var ongoingFetch = _this7.ongoingFetch(storeId, cacheKey);

        if (!_this7.isCacheStale(storeId, cacheKey) && cachedStore.size) {
          return resolve(cachedStore);
        } else {
          if (instantResolve) {
            resolve(_immutableContribCursor2['default'].from(_immutable2['default'].fromJS(store.toJSON())));
          }
          if (ongoingFetch) {
            return ongoingFetch.promise.then(function () {
              if (instantResolve) {
                return _this7.trigger('fetch:' + storeId, null, _this7.cached.cursor([storeId, cacheKey]));
              } else {
                return resolve(_this7.cached.cursor([storeId, cacheKey]));
              }
            });
          } else {
            var fetchPromise = store.fetch(fetchOptions)['catch'](function (err) {
              var allowedStatus = _this7.allowedStatusCodes.some(function (statusCode) {
                return statusCode === err.status;
              });
              if (allowedStatus) {
                var result = _immutableContribCursor2['default'].from(_immutable2['default'].fromJS(store.toJSON()));
                if (instantResolve) {
                  return _this7.trigger('fetch:' + storeId, null, result);
                } else {
                  return resolve(result);
                }
              } else {
                // reject for other statuses so error can be catched in router
                if (instantResolve) {
                  return _this7.trigger('fetch:' + storeId, err);
                } else {
                  return reject(err);
                }
              }
            });

            _this7.markFetchOngoing(storeId, cacheKey, fetchPromise);

            return fetchPromise.then(function () {
              var result = _this7.cached.cursor([storeId, cacheKey]).update(function (previousStore) {
                // TODO: figure out to a way to use merge here, it's problematic
                // with updates where List item is deleted
                // if (previousStore) {
                //   return previousStore.merge(store.toJSON());
                // } else {
                //   return Immutable.fromJS(store.toJSON());
                // }
                return _immutable2['default'].fromJS(store.toJSON());
              });

              if (instantResolve) {
                _this7.trigger('fetch:' + storeId, null, result);
              } else {
                resolve(result);
              }

              _this7.markCacheFresh(storeId, cacheKey);
              // wait 50ms before marking fetch completed to prevent multiple fetches
              // within small time window
              setTimeout(function () {
                _this7.markFetchCompleted(storeId, cacheKey);
              }, 50);
            });
          }
        }
      });
    }
  }, {
    key: 'fetchAll',
    value: function fetchAll() {
      var _this8 = this;

      var options = arguments[0] === undefined ? {} : arguments[0];

      var storeIds = Object.keys(options);
      return Promise.all(storeIds.map(function (storeId) {
        return _this8.fetch(storeId, options[storeId]);
      })).then(function (results) {
        return results.reduce(function (result, store, i) {
          result[storeIds[i]] = store;
          return result;
        }, {});
      });
    }
  }]);

  return Store;
})();

exports['default'] = Store;
module.exports = exports['default'];